/*
 * file Browser.java
 *
 * Licensed Materials - Property of IBM
 * Restricted Materials of IBM - you are allowed to copy, modify and 
 * redistribute this file as part of any program that interfaces with 
 * IBM Rational CM API.
 *
 * com.ibm.rational.stp.client.samples.Browser
 *
 *  Copyright IBM Corporation 2004, 2008.  All Rights Reserved.
 * Note to U.S. Government Users Restricted Rights:  Use, duplication or 
 * disclosure restricted by GSA ADP  Schedule Contract with IBM Corp.
 */
package com.ibm.rational.stp.client.samples;

import java.awt.BorderLayout;
import java.awt.Cursor;
import java.awt.FlowLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.io.File;
import java.io.FileOutputStream;
import java.util.Arrays;
import java.util.Comparator;
import java.util.List;

import javax.swing.DefaultComboBoxModel;
import javax.swing.JButton;
import javax.swing.JComboBox;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.wvcm.PropertyRequestItem;
import javax.wvcm.Resource;
import javax.wvcm.ResourceList;
import javax.wvcm.WvcmException;
import javax.wvcm.PropertyNameList.PropertyName;
import javax.wvcm.PropertyRequestItem.NestedPropertyName;
import javax.wvcm.PropertyRequestItem.PropertyRequest;

import com.ibm.rational.stp.client.samples.BrowserDataModel.Operations;
import com.ibm.rational.wvcm.stp.StpException;
import com.ibm.rational.wvcm.stp.StpProperty;
import com.ibm.rational.wvcm.stp.StpPropertyException;
import com.ibm.rational.wvcm.stp.StpProvider;
import com.ibm.rational.wvcm.stp.StpResource;
import com.ibm.rational.wvcm.stp.StpException.StpReasonCode;
import com.ibm.rational.wvcm.stp.StpProperty.MetaPropertyName;
import com.ibm.rational.wvcm.stp.cq.CqDbSet;
import com.ibm.rational.wvcm.stp.cq.CqProvider;
import com.ibm.rational.wvcm.stp.cq.CqUserDb;

/**
 * This is a SWING application that reads and displays the properties of a
 * resource using the generic Resource/Property interfaces of the CM API.
 * Starting from the name or type of a resource, a user can read and display
 * resource properties, following references from one resource to the next.
 * <p>
 * The user can start browsing either by entering the object selector for the
 * resource to be viewed or the user can select a resource type and ask the API
 * for all known folders containing that resource type.
 * <p>
 * In the latter case, the list of folders is displayed by
 * {@link Browser#showResourceList Browser.showResourceList}. By selecting an
 * entry in the display for a Resource, ResourceList or Property.List value, the
 * user can traverse to that object or list of objects and display them.
 * <p>
 * In the former case, the user types in the object selector for a specific
 * resource he is interested in, such as a specific record. All properties of
 * this record are then displayed by {@link Browser#showResource 
 * Browser.showResource} The Property.List property Record.ALL_FIELD_VALUES
 * contains all the schema-defined fields of the record and could be selected to
 * view those fields.
 * <p>
 * If a resource has content, that data can be displayed using
 * {@link BrowserDataModel#showContent BrowserDataModel.showContent()}.
 */
public class Browser {

    /**
     * Displays a list of folders in which are found resources of a specified
     * type or a resource at a given location.
     * 
     * @param name
     *            Either a String specifying the type of resource for which a
     *            folder list is to be generated or a String containing the
     *            location of the resource to be displayed. Must not be null.
     */
    static void showRoot(String name) {
        // We've had to defer allocation of the provider until we're here on the
        // event thread to satisfy the Apartment threading model used by COM.
        if (g_provider == null)
            try {
                g_provider = Utilities.getProvider();
            } catch (Exception ex) {
                Utilities.exception(null, "Provider", ex);
            }
        try {
            if (name.equals(CLEAR_QUEST_DB_SETS)) {
                showResourceList("ClearQuest Connections",
                                 g_provider.cqProvider().doGetDbSetList(PROPS),
                                 PROPS.toArray(),
                                 true);
            } else if (name.equals(CLEAR_CASE_WORKSPACES)) {
                showResourceList("ClearCase Workspaces",
                                 g_provider.ccProvider().getClientViewList(PROPS),
                                 PROPS.toArray(),
                                 true);
            } else {
                showResource("Resource " + name, (StpResource) g_provider
                    .resource(g_provider.location(name)), 
                    META_PROPS.toArray(), true);
            }
        } catch (Throwable ex) {
            Utilities.exception(null, "Show root", ex);
        }
    }
    
    /**
     * Displays the elements of a ResourceList in a table.
     * 
     * @param title
     *            The title for the window in which the resources are displayed.
     * @param resources
     *            The ResoureList containing the resources to be displayed.
     * @param properties
     *            A PropertyName array containing the PropertyName's for the
     *            properties to be displayed for each Resource. These properties
     *            should be defined in the proxies in the ResourceList.
     * @return The handle on the window in which the resource list is displayed.
     */
    static JFrame showResourceList(String title,
                                   final ResourceList<? extends StpResource> resources,
                                   final NestedPropertyName<?>[] properties,
                                   final boolean validOnly) {
        // Generate the data model for displaying the resource list,
        // one resource per row, one property per column
        BrowserDataModel dataModel = new BrowserDataModel() {
            public int getColumnCount() {
                return 1+properties.length;
            }

            public int getRowCount() {
                return resources.size();
            }

            public Object getValueAt(int row, int col) {
                if (col == 0)
                    return Utilities.resourceType(resources.get(row));
                
                try {
                    return resources.get(row).getProperty(properties[col - 1]
                        .getRoot());
                } catch (Throwable ex) {
                    // If an object name selector was to be displayed
                    // substitute the resource location if that property
                    // is unavailable. (Not all resources have them.)
                    return StpResource.USER_FRIENDLY_LOCATION
                        .equals(properties[col-1]) ? resources.get(row)
                        : exceptionImage(ex, resources.get(row));
                }
            }

            public String getColumnName(int col) {
                return col==0? "resource-type": properties[col-1].getRoot().getName();
            }

            // The only showable objects are Resource's, so use showResource()
            void show(Object viewable) throws WvcmException {
                StpResource res = (StpResource) viewable;

                try {
                    m_frame.setCursor(Cursor
                        .getPredefinedCursor(Cursor.WAIT_CURSOR));
                    showResource(Utilities.resourceType(res) + " "
                        + res.location().string(), res, 
                        META_PROPS.toArray(), validOnly);
                } finally {
                    m_frame.setCursor(Cursor
                        .getPredefinedCursor(Cursor.DEFAULT_CURSOR));
                }
            }

            // Since each row represents a resource, each row is always
            // showable
            Object getViewable(int row) {
                return resources.get(row);
            }

            private static final long serialVersionUID = 1L;
        };

        return dataModel.showModel(title, false);
    }

    /**
     * Retrieves all properties of a given resource and displays them in a
     * table.
     * 
     * @param title
     *            The title for the window in which the resource is displayed.
     * @param res
     *            A Resource proxy for the resource to be displayed.
     * @param metaProperties
     *            A MetaPropertyName[] containing the meta-properties to be
     *            displayed for each property of the resource.
     * @return A handle for the window containing the display.
     * @throws WvcmException
     *             if resource denoted by the given Resource proxy can not be
     *             read.
     */
    static JFrame showResource(String title,
                               StpResource res,
                               final PropertyRequestItem.NestedPropertyName<?>[] metaProperties,
                               boolean validOnly) throws WvcmException {
        return showPropertyList(title,
                                res,
                                StpResource.ALL_PROPERTIES,
                                metaProperties,
                                validOnly);
    }

    /**
     * Retrieves the value of a Property.List-valued property from a given
     * resource and displays the results in a table. The display includes an
     * option to view the content of the resource as well as the resource or
     * resource list referenced by any of the displayed properties.
     * 
     * @param title
     *            The title for the window in which the property list is
     *            displayed
     * @param resource
     *            A proxy for the Resource whose properties are to be displayed.
     * @param property
     *            A PropertyName specifying the Property.List-valued property of
     *            the resource that is to be retrieved and displayed.
     * @param metaProperties
     *            The meta-properties of each Property on the list that are to
     *            be retrieved and displayed.
     * @return A handle for the window in which the Property.List is displayed.
     * @throws WvcmException
     *             If the specified resource or property cannot be retrieved
     *             from the repository.
     */
    static JFrame showPropertyList(final String title,
                                   final StpResource resource,
                                   final PropertyName<StpProperty.List<StpProperty<?>>> property,
                                   final NestedPropertyName<?>[] metaProperties,
                                   final boolean validOnly)
                    throws WvcmException {
        PropertyRequest wantedProps =
            new PropertyRequest(property.nest(StpProperty.VALUE
                .nest(metaProperties)));
        final StpResource res = (StpResource) resource.doReadProperties(wantedProps);
        final StpProperty.List<?> total = res.getProperty(property);
        final Operations operations = getOperations(res);

        // Filter out properties that could not be read.
        final StpProperty.List<StpProperty<?>> valid = 
            new StpProperty.List<StpProperty<?>>();

        for(StpProperty<?> prop: total) {
            try {
                // If getValue() doesn't blow up, we can display the property
                prop.getValue();
                valid.add(prop);
            } catch (WvcmException ex) {
            }
        }

        // Sort the properties by their simple name
        StpProperty<?>[] props;

        if (validOnly)
            props = (StpProperty[]) valid.toArray(new StpProperty[valid.size()]);
        else
            props = (StpProperty[]) total.toArray(new StpProperty[total.size()]);

        Arrays.sort(props, new Comparator<StpProperty<?>>() {
            public int compare(StpProperty<?> arg0, StpProperty<?> arg1) {
                return arg0.getPropertyName().getName()
                        .compareTo(arg1.getPropertyName().getName());
            }
        });

        final StpProperty.List<StpProperty<?>> properties = 
                                        new StpProperty.List<StpProperty<?>>();

        properties.addAll(Arrays.asList(props));

        // Build the data model for displaying one property per row, one
        // meta-property per column.
        BrowserDataModel dataModel = new BrowserDataModel() {
            public int getColumnCount() {
                return metaProperties.length;
            }

            public int getRowCount() {
                return properties.size();
            }

            public Object getValueAt(int row, int col) {
                try {
                    return ((StpProperty<?>) properties.get(row))
                        .getMetaProperty((MetaPropertyName<?>)
                                         metaProperties[col].getRoot());
                } catch (Throwable ex) {
                    if (StpProperty.TYPE.equals(metaProperties[col]))
                        try {
                            return typeImage(((StpProperty<?>) properties.get(row))
                                .getValue());
                        } catch (WvcmException e) {
                        }

                    return exceptionImage(ex, properties.get(row));
                }
            }

            public String getColumnName(int col) {
                return metaProperties[col].getRoot().getName();
            }

            /**
             * Properties whose values are resources or lists of resources,
             * properties ChildBinding objects or ParentBinding objects are
             * viewable.
             */
            Object getViewable(int row) {
                StpProperty<?> result = properties.get(row);
                try {
                    Object val = result.getValue();

                    if (val != null
                        && (val instanceof StpResource 
                            || Utilities.isListOfResources(val)
                            || val instanceof StpProperty.List))
                        return result;
                } catch (WvcmException ex) {
                }

                return null;
            }

            /**
             * Generates a new table window for a resource, resource list or
             * property list.
             */
            void show(Object viewable) throws WvcmException {
                StpProperty<?> property = (StpProperty<?>) viewable;
                Object val = property.getValue();

                try {
                    m_frame.setCursor(Cursor
                        .getPredefinedCursor(Cursor.WAIT_CURSOR));

                    if (Utilities.isListOfResources(val)) {
                        showResourceList(property.getName() + " of "
                                             + Utilities.resourceType(res)
                                             + " " + res.location().string(),
                                         Utilities.toResourceList(res.stpProvider(), 
                                                                  (List<?>)val),
                                         PROPS.toArray(),
                                         validOnly);
                    } else if (val instanceof StpResource) {
                        StpResource res = (StpResource) val;
                        showResource(Utilities.resourceType(res) + " "
                                         + res.location().string(),
                                     res,
                                     metaProperties,
                                     validOnly);
                    } else if (val instanceof StpProperty.List) {
                        PropertyName<StpProperty.List<StpProperty<?>>> prop =
                            StpException
                                .<PropertyName<StpProperty.List<StpProperty<?>>>> 
                                    unchecked_cast(property
                                                       .getPropertyName());
                        
                        showPropertyList(property.getName() + " of "
                                             + Utilities.resourceType(res)
                                             + " " + res.location().string(),
                                         res,
                                         prop,
                                         metaProperties,
                                         validOnly);
                    }
                } finally {
                    m_frame.setCursor(Cursor
                        .getPredefinedCursor(Cursor.DEFAULT_CURSOR));
                }
            }

            String toggleErrorsLabel() {
                return m_validPropertyCount == m_totalPropertyCount ? ""
                    : properties.size() == m_validPropertyCount ? "Show Errors"
                        : "Hide Errors";
            }

            void toggleErrors() throws WvcmException {
                try {
                    m_frame.setCursor(Cursor
                        .getPredefinedCursor(Cursor.WAIT_CURSOR));
                    showPropertyList(title,
                                     resource,
                                     property,
                                     metaProperties,
                                     !validOnly);
                } finally {
                    m_frame.dispose();
                }
            }

            void redisplay() throws WvcmException {
                try {
                    m_frame.setCursor(Cursor
                        .getPredefinedCursor(Cursor.WAIT_CURSOR));

                    showPropertyList(title,
                                     resource,
                                     property,
                                     metaProperties,
                                     validOnly);
                } finally {
                    m_frame.dispose();
                }
            }

            /**
             * Generates a display for the content of this resource
             */
            void showContent() throws Throwable {
                File temp = File.createTempFile("browser", "tmp");
                FileOutputStream stream = new FileOutputStream(temp);

                res.doReadContent(stream, null);
                showFile("Content of " + Utilities.resourceType(res) + " "
                    + res.location().string(), temp);
            }

            /**
             * Returns the Operations object associated with this window.
             * @see CcBrowser
             */
            Operations getOperationsObject() {
                return operations;
            }

            private static final long serialVersionUID = 1L;

            /** The total number of properties defined by the resource proxy */
            private int m_totalPropertyCount = total.size();

            /** The total number of valid properties defined by the proxy */
            private int m_validPropertyCount = valid.size();
        };

        return dataModel.showModel(title, true);
    }

    /**
     * Starts browsing at a user-supplied resource location or in the folders
     * designated for a user-supplied type of resource.
     * 
     * @param args
     *            Not used in this application
     */
    public static void main(String[] args) throws Exception {
        /**
         * Present the list to the user and marshal requests for folder lists
         * and specific resources to the showRoot() method.
         */
        String help = "Enter a Location or Select a resource category and Click 'Browse'";
        String prompt = "<type>.<namespace>:<name>@<repository>";

        final JFrame frame = new JFrame("Resource Browser");
        final JPanel panel = new JPanel(new BorderLayout());
        final JPanel subpanel = new JPanel(new FlowLayout());
        final DefaultComboBoxModel model = new DefaultComboBoxModel();

        // Add object selectors for sample databases. Edit this code to match
        // installed resources as desired.
        model.insertElementAt("cq.record:Defect/SAMPL00000005@7.0.0/SAMPL", 0);
        model.insertElementAt(CLEAR_CASE_WORKSPACES, 0);
        model.insertElementAt(CLEAR_QUEST_DB_SETS, 0);
        model.insertElementAt(prompt, 0);
        model.setSelectedItem(prompt);

        final JComboBox box = new JComboBox(model);

        panel.add(box, BorderLayout.CENTER);
        box.setEditable(true);

        final JButton browse = new JButton("Browse");

        subpanel.add(browse, BorderLayout.SOUTH);
        browse.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent arg0) {
                String selection = (String) box.getSelectedItem();

                if (model.getIndexOf(selection) < 0)
                    model.insertElementAt(selection, 0);

                try {
                    frame.setCursor(Cursor
                        .getPredefinedCursor(Cursor.WAIT_CURSOR));

                    showRoot(selection);
                } finally {
                    frame.setCursor(Cursor
                        .getPredefinedCursor(Cursor.DEFAULT_CURSOR));
                }
            }
        });

        JButton exit = new JButton("Exit");
        subpanel.add(exit);
        exit.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent arg0) {
                System.exit(0);
            }
        });

        final JButton cqurl = new JButton("URL...");
        subpanel.add(cqurl);
        cqurl.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent arg0) {
                getUrl("http://qwin115:9082/TeamWeb/services/Team");
        }});

        panel.add(new JLabel(help), BorderLayout.NORTH);
        panel.add(subpanel, BorderLayout.SOUTH);
        frame.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
        frame.setContentPane(panel);
        frame.setBounds(200, 200, 500, 110);
        frame.setVisible(true);
    }

    static boolean getUrl(String defaultURL)
    {
        try {
            if (g_provider == null)
                    g_provider = Utilities.getProvider();

            String url = g_provider.getServerUrl();
            
            if (url == null || url.length() == 0)
                url =defaultURL;
            
            url =
                JOptionPane.showInputDialog("Enter a URL for the server", url);
            
            g_provider.setServerUrl(url);
            return true;
        } catch (Exception ex) {
            Utilities.exception(null, "Provider", ex);
        }
        
        return false;
    }
    /**
     * Generates a short string to identify a given throwable
     * 
     * @param ex
     *            A Throwable that is to be identified.
     * @param obj
     *            The object that promulgated the exception.
     * @return A short String providing, in a concise format, identification of
     *         and relevant information from the throwable.
     */
    static Object exceptionImage(Throwable ex, Object obj) {
        StpReasonCode code = null;
        String prefix = "";
        
        if (ex instanceof StpException) {
            code = ((StpException) ex).getStpReasonCode();
        } else if (obj instanceof StpPropertyException) {
            StpPropertyException pe = (StpPropertyException) obj;
            PropertyName<?> name = pe.getPropertyName();

            code = pe.getStpReasonCode();

            if (name != null)
                prefix = name.getName() + ": ";
        } else {
            String name = ex.getClass().getName();
            return name.substring(name.lastIndexOf(".") + 1);
        }
        
        // Look through the wrapper to display more information
        if (code == StpReasonCode.PROPERTY_RETRIEVAL_FAILED)
            code = ((StpException)ex.getCause()).getStpReasonCode();
        
        return prefix + code;
    }

    /**
     * Generates the unqualified name of the type of a given object
     * 
     * @param obj
     *            The object whose value is to be displayed. May be null.
     * @return The leaf name of the type of the object.
     */
    static Object typeImage(Object obj) {
        if (obj == null)
            return "(null)";

        String name = obj.getClass().getName();
        int dot = name.lastIndexOf('.');

        return dot < 0 ? name : name.substring(dot + 1);
    }

    /**
     * If the g_operationsClass variable is set, returns a new instance of that
     * class initialized with the given Resource proxy.
     * 
     * @param res
     *            A Resource proxy. Must not be null.
     * @return A new Operations object if the operations class has been assigned
     *         to g_operationsClass; otherwise null.
     */
    private static Operations getOperations(StpResource res) {
        if (g_operationsClass != null)
            try {
                return (Operations) g_operationsClass
                    .getConstructor(new Class[] { StpResource.class })
                    .newInstance(new Object[] { res });
            } catch (Throwable t) {
                Utilities.exception(null, "getOperations", t);
            }

        return null;
    }

    private static final MetaPropertyName<Object> VALUE_AS_OBJECT = 
                new MetaPropertyName<Object>(StpProperty.VALUE.getNamespace(),
                                             StpProperty.VALUE.getName());

    /**
     * Properties to be displayed for elements of a ResourceList Also the
     * properties requested for resource-valued properties.
     */
    private static final PropertyRequest PROPS = new PropertyRequest(
                    StpResource.USER_FRIENDLY_LOCATION  /*,
                        StpResource.COMMENT*/ );

    /**
     * Meta-properties to be displayed for each property Also the properties
     * requested for property-valued properties
     */
    static final PropertyRequest META_PROPS = 
        new PropertyRequest( new PropertyRequestItem[] { 
                                StpProperty.NAME,
                                StpProperty.TYPE, 
                                StpProperty.VALUE.nest(PROPS)});

    /** Drop-down entry for requesting ClearQuest connections */
    private static final String CLEAR_QUEST_DB_SETS = "<ClearQuest Connections>";

    /** Drop-down entry for requesting local ClearCase views */
    private static final String CLEAR_CASE_WORKSPACES = "<Local ClearCase Views>";

    /** 
     * The class to use for Operations if supported by the application using
     * the Browser class.
     * 
     *  @see CcBrowser#main(String[])
     */
    static Class<?> g_operationsClass = null;

    /** The Provider instance used by all instances of the Browser in a process */
    private static StpProvider g_provider = null;
}
