/*
 * file QueryUtilities.java
 * 
 * Licensed Materials - Property of IBM
 * Restricted Materials of IBM - you are allowed to copy, modify and 
 * redistribute this file as part of any program that interfaces with 
 * IBM Rational CM API.
 *
 * com.ibm.rational.stp.client.samples.QueryUtilities
 *
 *  Copyright IBM Corporation 2004, 2008.  All Rights Reserved.
 * Note to U.S. Government Users Restricted Rights:  Use, duplication or 
 * disclosure restricted by GSA ADP  Schedule Contract with IBM Corp.
 */
package com.ibm.rational.stp.client.samples;

import java.util.ArrayList;
import java.util.EnumSet;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;

import javax.wvcm.ResourceList;
import javax.wvcm.WvcmException;
import javax.wvcm.PropertyRequestItem.PropertyRequest;

import com.ibm.rational.wvcm.stp.StpException;
import com.ibm.rational.wvcm.stp.StpLocation;
import com.ibm.rational.wvcm.stp.cq.CqFieldDefinition;
import com.ibm.rational.wvcm.stp.cq.CqFieldValue;
import com.ibm.rational.wvcm.stp.cq.CqQuery;
import com.ibm.rational.wvcm.stp.cq.CqRecordType;
import com.ibm.rational.wvcm.stp.cq.CqQuery.DisplayField;
import com.ibm.rational.wvcm.stp.cq.CqQuery.DisplayField.SortType;
import com.ibm.rational.wvcm.stp.cq.CqQuery.Filter.Operation;
import com.ibm.rational.wvcm.stp.cq.CqQuery.FilterLeaf.TargetType;

/**
 * A CmdBaseQuery
 */
public abstract class QueryUtilities {
    static private enum Kind {
        /** FilterSym.kind for an operation with no operands */
        NILARY, 
        /** FilterSym.kind for an operation with one operand */
        UNARY, 
        /** FilterSym.kind for an operation with two operands */
        BINARY, 
        /** FilterSym.kind for an operation with one or more operands */
        VARIADIC, 
        /** FilterSym.kind for target specification token */
        TARGET, 
        /** FilterSym.kind for a target separator token */
        SEPARATOR, 
        /** FilterSym.kind for the end of a FilterLeaf expression */
        END
    };

    /** Kind mask for target specification token */
    private static final EnumSet<Kind> TARGET_MASK = EnumSet.of(Kind.TARGET);

    /** Kind mask for a target separator token */
    private static final EnumSet<Kind> SEPARATOR_MASK = EnumSet.of(Kind.SEPARATOR);

    /** Kind mask for the end of a FilterLeaf expression */
    private static final EnumSet<Kind> END_MASK = EnumSet.of(Kind.END);

    
    /**
     * The representation for a symbol in a filtering expression
     */
    private static class FilterSym<T> {
        /** The Operation or TargetType for this filter symbol */
        T code;

        /** The token image used to represent this filter symbol */
        String symbol;

        /** The kind of this symbol */
        Kind kind;

        /**
         * Creates a new FilterSym object for an Operation.
         * 
         * @param o The Operation enumerator
         * @param i The Operation symbol
         * @param t The kind code
         */
        private FilterSym(T o, String i, Kind t)
        {
            code = o;
            symbol = i;
            kind = t;
        }

        /**
         * Creates a new FilterSym object for a TargetType enumerator.
         * 
         * @param o The TargetType code
         * @param i The target name
         */
        private FilterSym(T o, String i)
        {
            code = o;
            symbol = i;
            kind = Kind.TARGET;
        }

        /**
         * Creates a new FilterSym object for a target token.
         * 
         * @param token The token string.
         */
        private FilterSym(String token)
        {
            symbol = token;
            kind = Kind.TARGET;
            code = StpException.<T>unchecked_cast(TargetType.CONSTANT);
//
//            if (token.startsWith("cq.")
//                || token.startsWith("record:")) {
//                code = StpException.<T>unchecked_cast(TargetType.REFERENCE);
//            }
        }

        /**
         * Creates a new FilterSym object for a parameterized special token
         * such as TargetType.PROMPTED
         * 
         * @param sym The (parameterless) FilterSym for the special token
         * @param param The parameter for the special token.
         */
        private FilterSym(FilterSym<T> sym, String param)
        {
            code = sym.code;
            symbol = param;
            kind = sym.kind;
        }
    }

    /** Pre-defined Filter token definitions */
    private static FilterSym[] symbol = {
        new FilterSym<Operation>(Operation.CONJUNCTION, "[and]", Kind.SEPARATOR),
        new FilterSym<Operation>(Operation.DISJUNCTION, "[or]", Kind.SEPARATOR),
        new FilterSym<Operation>(Operation.IS_EQUAL, "[eq]", Kind.UNARY),
        new FilterSym<Operation>(Operation.IS_NOT_EQUAL, "[ne]", Kind.UNARY),
        new FilterSym<Operation>(Operation.IS_LESS_THAN, "[lt]", Kind.UNARY),
        new FilterSym<Operation>(Operation.IS_LESS_THAN_OR_EQUAL, "[le]", Kind.UNARY),
        new FilterSym<Operation>(Operation.IS_GREATER_THAN, "[gt]", Kind.UNARY),
        new FilterSym<Operation>(Operation.IS_GREATER_THAN_OR_EQUAL, "[ge]", Kind.UNARY),
        new FilterSym<Operation>(Operation.HAS_SUBSTRING, "[like]", Kind.UNARY),
        new FilterSym<Operation>(Operation.HAS_SUBSTRING, "[has]", Kind.UNARY),
        new FilterSym<Operation>(Operation.HAS_SUBSTRING, "[contains]", Kind.UNARY),
        new FilterSym<Operation>(Operation.HAS_NO_SUBSTRING, "[unlike]", Kind.UNARY),
        new FilterSym<Operation>(Operation.HAS_NO_SUBSTRING, "[lacks]", Kind.UNARY),
        new FilterSym<Operation>(Operation.IS_BETWEEN, "[between]", Kind.BINARY),
        new FilterSym<Operation>(Operation.IS_NOT_BETWEEN, "[outside]", Kind.BINARY),
        new FilterSym<Operation>(Operation.IS_NULL, "[null]", Kind.NILARY),
        new FilterSym<Operation>(Operation.IS_NOT_NULL, "[set]", Kind.NILARY),
        new FilterSym<Operation>(Operation.IS_IN_SET, "[is]", Kind.VARIADIC),
        new FilterSym<Operation>(Operation.IS_NOT_IN_SET, "[isnt]", Kind.VARIADIC),
        new FilterSym<TargetType>(TargetType.USER, "[user]"),
        new FilterSym<TargetType>(TargetType.USER, "[me]"),
        new FilterSym<TargetType>(TargetType.YESTERDAY, "[yesterday]"),
        new FilterSym<TargetType>(TargetType.TODAY, "[today]"),
        new FilterSym<TargetType>(TargetType.TOMORROW, "[tomorrow]"),
        new FilterSym<TargetType>(TargetType.DATE_ONLY, "[date]"),
        new FilterSym<TargetType>(TargetType.DATE_TIME, "[time]"),
        new FilterSym<TargetType>(TargetType.PROMPTED, "[prompt:]"),
//        new FilterSym<TargetType>(TargetType.FIELD, "[field:]"),
        null };

    /**
     * Tokenizes a stream containing a filtering expression from the command
     * line
     */
    private static class Tokenizer {
        /**
         * Creates a new FilterTokenizer object based on a given stream.
         * 
         * @param stream The String to be tokenized
         */
        Tokenizer(String stream)
        {
            m_stream = stream;
            m_text = stream;
        }

        /**
         * Fetches the next token in the stream, verifies that its kind
         * satisfies the mask parameter; and, if specified, verifies that its
         * code matches the match parameter, and finally, returns its FilterSym
         * structure;
         * 
         * @param mask A bit mask specifying the type(s) of token that are
         *            acceptable return values.
         * @param match A specific FilterSym.code value that the token is
         *            required to match; may be null if no specific code match
         *            is required.
         * @return A FilterSym object representing the next token in the stream.
         */
        <T> FilterSym<T> next(EnumSet<Kind> mask, Object match)
        {
            String token = null;

            if (m_stream == null || m_stream.length() == 0) {
                // End of stream; result is zero
                if (mask.contains(Kind.END)) {
                    return null;
                }

                throw new IllegalArgumentException("Unexpected end of filter in '"
                    + m_text + "'");
            } else {
                // All blanks in the stream are significant
                // A token is either text enclosed in []
                // or the longest span of text containing no '['
                int lb = m_stream.indexOf("[");

                if (lb < 0) {
                    // The last token in the stream; a Target
                    token = m_stream;
                    m_stream = null;

                    if (mask.contains(Kind.TARGET)) {
                        return new FilterSym<T>(token);
                    }

                    throw new IllegalArgumentException("Unexpected symbol '"
                        + token + "' in filter '" + m_text + "'");
                } else if (lb == 0) {
                    // Have a special token [...]
                    int rb = m_stream.indexOf("]");

                    if (rb < 0) {
                        throw new IllegalArgumentException("Special token '"
                            + m_stream + "' has no closing ']' in '"
                            + m_text + "'");
                    }

                    token = m_stream.substring(0, rb + 1); // includes "[" and
                    // "]"
                    m_stream = m_stream.substring(rb + 1);

                    int colon = token.indexOf(":");
                    String arg = null;

                    if (colon > 0) {
                        // Elide argument for lookup purposes
                        arg = token.substring(colon + 1, token.length()-1);
                        token = token.substring(0, colon) + ":]";
                    }

                    // Look up special token in symbol table

                    for (FilterSym<T> result: symbol) {
                        if (result.symbol.equals(token)) {
                            if (mask.contains(result.kind)
                                && (match == null || match.equals(result.code))) {
                                if (arg != null) {
                                    return new FilterSym<T>(result, arg);
                                } else {
                                    return result;
                                }
                            }

                            throw new IllegalArgumentException("Unexpected token '"
                                + token + "' in '" + m_text + "'");
                        }
                    }

                    // Symbol not found in symbol table
                    throw new IllegalArgumentException("Undefined token '"
                        + token + "' in '" + m_text + "'");
                } else {
                    // Not a special symbol; a TARGET token
                    token = m_stream.substring(0, lb);
                    m_stream = m_stream.substring(lb);

                    if (mask.contains(Kind.TARGET)) {
                        return new FilterSym<T>(token);
                    }

                    throw new IllegalArgumentException("Unexpected symbol '"
                        + token + "' in '" + m_text + "'");
                }
            }
        }

        /**
         * Fetches the next token in the stream, verifies that its kind
         * satisfies the mask parameter and finally, returns its FilterSym
         * structure;
         * 
         * @param mask A bit mask specifying the type(s) of token that are
         *            acceptable return values.
         * 
         * @return A FilterSym object representing the next token in the stream.
         */
        FilterSym next(EnumSet<Kind> mask)
        {
            return next(mask, null);
        }

        FilterSym<TargetType> nextTarget()
        {
            return next(TARGET_MASK, null);
        }
        
        FilterSym<Operation> nextOperation()
        {
            return next(EnumSet.of(Kind.NILARY, Kind.UNARY, Kind.BINARY, Kind.VARIADIC), 
                        null);
        }

        /**
         * Fetches the next token in the stream, without verification, and
         * returns its FilterSym structure;
         * 
         * @return A FilterSym object representing the next token in the stream.
         */
        FilterSym next()
        {
            return next(EnumSet.allOf(Kind.class));
        }

        /** The remainder of the character stream from which tokens are fetched */
        private String m_stream;

        /** The original input stream */
        private String m_text;
    };

    /**
     * Locates that CqFieldDefinition that corresponds to a given field name.
     * Errors or warnings are generated if the given name doesn't match a
     * defined field exactly.
     *
     * @param  name  The field name to be matched. Must not be null.
     * @param  fields  The list of FieldDefinitions against which the name is to
     *                 be compared.
     *
     * @return  The CqFieldDefinition that matches the given name; <b>null</b>
     *          returned if no match is found.
     *
     * @throws  WvcmException  Thrown if the supplied FieldDefinitions do not
     *                         define the DISPLAY_NAME property.
     */
    private static CqFieldDefinition findField(String name, ResourceList fields)
        throws WvcmException
    {
        CqFieldDefinition fieldDef = null;

        for (int i = 0; i < fields.size(); ++i) {
            fieldDef = (CqFieldDefinition) fields.get(i);
            String fieldName = fieldDef.getDisplayName();

            if (name.compareToIgnoreCase(fieldName) == 0) {
                if (!name.equals(fieldName))
                    System.out.println("Field '" + name
                        + "' should be spelled " + fieldName);

                return fieldDef;
            }
        }

        throw new IllegalArgumentException("'" + name
            + "' does not name a field of record type "
            + fieldDef.getRecordType().getDisplayName());
    }

    /**
     * Constructs a CqFieldDefinition[] from a field path specification
     * consisting of dotted segments: field1.field2. ... .fieldN
     * 
     * @param query A CqQuery proxy for the query in which the field path will
     *            be used. Must define
     *            PRIMARY_RECORD_TYPE.nest(FIELD_DEFINITIONS.nest(FIELD_DEFINITION_PROPERTIES))
     * @param path A dot-separated list of field names describing the path
     * from a field of the primary resource type, through references, to the
     * targeted field.
     * @return A CqFieldDefinition[] containing a valid CqFieldDefinition proxy 
     * for each segment of the path.
     * @throws WvcmException If the information needed to analyze the path is
     * unavailable or unobtainable.
     */
    private static CqFieldDefinition[] buildFieldPath(CqQuery query, String path)
        throws WvcmException
    {
        String[] segments = path.split("\\.");
        CqFieldDefinition[] result = new CqFieldDefinition[segments.length];
        CqRecordType recType = (CqRecordType)query.getPrimaryRecordType();

        for (int i = 0; i < segments.length; ++i) {
            String fieldname = segments[i];

            if (recType == null)
                throw new IllegalArgumentException
                    ("No record type information available for field path segment "
                         + fieldname);

            CqFieldDefinition def = findField(fieldname, 
                                              recType.getFieldDefinitions());

            recType = getReferencedRecordType(result[i] = def);
        }

      return result;
    }

    /**
     * Returns a fully-populated proxy for the record type of the records
     * referenced by a given field definition.
     * 
     * @param def The CqFieldDefinition whose referenced record type is desired.
     *            must not be null.
     * @return A fully-populated CqRecordType proxy for the record type of the
     *         records referenced by the values of the given field definition;
     *         null if the field value doesn't reference a record (or record
     *         list). The results are cached to avoid unnecessary interactions
     *         with the serve.
     * @throws WvcmException
     */
    private static CqRecordType getReferencedRecordType(CqFieldDefinition def)
    throws WvcmException
    {   
        if (def.getFieldType() != CqFieldValue.ValueType.RESOURCE
                && def.getFieldType() != CqFieldValue.ValueType.RESOURCE_LIST)
            return null;
            
        CqRecordType recType = def.getReferencedRecordType();
        CqRecordType result = (CqRecordType) 
            g_recordTypeMap.get(recType.getUserFriendlyLocation());
        
        if (result == null) {
            result = (CqRecordType) recType
                .doReadProperties(RECORD_TYPE_WITH_FIELDS);
            g_recordTypeMap.put(result.getUserFriendlyLocation(), result);
        }
        
        return result;
    }

    /**
     * A Map from user-friendly location to fully-populated record type proxy.
     * Used to avoid unnecessary trips to the server to get this static
     * information
     */
    private static HashMap<StpLocation, CqRecordType> g_recordTypeMap = 
        new HashMap<StpLocation, CqRecordType>();
    
    /**
     * Verifies that the operation in the input is consistent with the previous
     * operations used.
     * 
     * @param oldOp The previous Operation used; may be null if this is first
     *            use of an Operation.
     * @param newOp The current Operation in the input.
     * @return The newOp.
     * @throws RuntimeException if the newOp is not compatible with the oldOp.
     */
    private static Operation checkOp(Operation oldOp, Operation newOp)
    {
        if (oldOp != null && !oldOp.equals(newOp))
            throw new RuntimeException("Unexpected operation '" + newOp
                + "' in filtering expression");
        
        return newOp;
    }

    /**
     * Parses a single filter leaf specification and construct the corresponding
     * FilterLeaf structure for it. The specification is in the form
     * <i>field-path</i>[op]<i>target-list</i> Example specifications are...
     * 
     * <pre>
     *     owner[eq][user]
     *     date[between]10/26/94[and]10/24/95
     *     submit_date[outside]8-feb-02[and]10-mar-02
     *     owner.name[unlike]Fred
     *     State[is]Submitted[or]Closed
     *     Severity[isnt]High[or]Low
     *     Description[null]
     *     Priority[set]
     * </pre>
     * 
     * @param query The Query proxy for the query that will be using the filter
     * @param image String containing the filter leaf expression.
     * 
     * @return A Query.FilterLeaf structure for the filter leaf expression
     * 
     * @throws WvcmException If the necessary information is not available or
     *             attainable from the server.
     */
    protected static CqQuery.FilterLeaf parseFilterLeaf(CqQuery query, 
                                                String image)
        throws WvcmException
    {
        Tokenizer token = new Tokenizer(image);
        FilterSym<TargetType> field = token.nextTarget();
        FilterSym<Operation> op = token.nextOperation();
        ArrayList<FilterSym<TargetType>> targets = new ArrayList<FilterSym<TargetType>>();
    
        switch (op.kind) {
        case NILARY: // field op
            token.next(END_MASK);
            break;
    
        case UNARY: // field op target
            targets.add(token.nextTarget());
            token.next(END_MASK);
            break;
    
        case BINARY: // field op target [and] target
            targets.add(token.nextTarget());
            token.next(SEPARATOR_MASK, Operation.CONJUNCTION);
            targets.add(token.nextTarget());
            token.next(END_MASK);
            break;
    
        case VARIADIC: // field op target [or] target [or] target ...
            targets.add(token.nextTarget());
    
            for (;;) {
                if (token.next(EnumSet.of(Kind.SEPARATOR, Kind.END),
                               Operation.DISJUNCTION) == null) {
                    break;
                }
    
                targets.add(token.nextTarget());
            }
    
            break;
        }
    
        CqFieldDefinition[] path = buildFieldPath(query, field.symbol);
        CqQuery.FilterLeaf result = query.cqProvider().buildFilterLeaf(path, op.code);
    
        for (FilterSym<TargetType> target: targets)
            result.addTarget(target.code, target.symbol);
    
        return result;
    }
    
    /**
     * Parses the specification for the parameter value that matches a dynamic
     * filter.
     * 
     * @param query The query for which the parameter is being specified
     * @param n The index of the dynamic filter in the query's dynamic filter
     *            list.
     * @param param The parameter value specification, which is a filter leaf
     *            specification without the field path and optionally without
     *            the operator, these being defaulted from the dynamic filter.
     *            Must not be <b>null</b>, but may be "" to indicate no value
     *            is to be supplied for the dynamic filter
     * @return A FilterLeaf object representing the parameter operation and 
     * values to be used for the n-th dynamic filter of the query as specified
     * by the parameter string.
     * @throws WvcmException
     */
    protected static CqQuery.FilterLeaf parseDynamicFilter(CqQuery query, 
                                                   int n, 
                                                   String param)
    throws WvcmException
    {
        if (param.equals(""))
            return null;
        
        CqQuery.FilterLeaf filter = query.getDynamicFilters()[n];
        
        if (!param.startsWith("[")) {
            Operation op = filter.getOperation();
            
            for (int i=0; i < symbol.length; ++i)
                if (symbol[i].code == op) {
                    param = symbol[i].symbol + param;
                    break;
                }
        }
        
        return parseFilterLeaf(query, filter.getSourceName() + param);
    }

    /**
     * Parses a filtering expression, builds a Filter structure for it, and
     * defines that as the value of the query's FILTERING property. The
     * filtering expression must conform to the following grammar for a filter
     * 
     * <pre>
     * filter ::= filterLeaf
     * filter ::= [(] filter [)]
     * filter ::= filter ([or] filter)+
     * filter ::= filter ([and] filter)+
     * </pre>
     * 
     * Examples
     * <ul>
     * <li>name[eq]Fred
     * <li>name[eq]Fred [and] owner[ne]George
     * <li>name[eq]Fred [or] owner[ne]George
     * <li>name[eq]Fred [and] status[isnt]fired[or]dead [and] owner[ne]George
     * <li>name[eq]Fred [or] submit_date[between]1/1/2001[and]3/3/2003 [or]
     * owner[ne]George
     * <li>name[eq]Fred [or] [(] submit_date[between]1/1/2001[and]3/3/2003
     * [and] owner[ne]George [)]
     * <li>name[eq]Fred [or] [(] submit_date[between]1/1/2001[and]3/3/2003
     * [and] reason[null] [)] [or] owner[ne]George]
     * <li>[(] submit_date[between]1/1/2001[and]3/3/2003 [and] reason[null] [)]
     * [or] name[eq]Fred [or] owner[ne]George]
     * </ul>
     * 
     * @param query The CqQuery proxy in which the FILTERING property is to be
     *            defined. This proxy must define the PRIMARY_RESOURCE property
     *            using a CqRecordType proxy that defines the FIELD_DEFINITIONS 
     *            property.
     * @param filterItems An Iterator over String objects producing the terms of
     *            the filtering expression in order left to right, where a term
     *            is one of the elements, <code>filterLeaf</code>,
     *            <code>[(]</code>, <code>[)]</code>, <code>[and]</code>,
     *            or <code>[or]</code> in the above grammar
     * @return A Filter object representing the specified filtering expression
     * @throws WvcmException
     */
    protected static CqQuery.Filter parseFilter(CqQuery query, Iterator filterItems)
        throws WvcmException
    {
        List<CqQuery.Filter> operands = new ArrayList<CqQuery.Filter>();
        Operation op = null;

        String term = (String) filterItems.next();

        if (term.equals("[(]")) {
            operands.add(parseFilter(query, filterItems));
        } else
            operands.add(parseFilterLeaf(query, term));
        
        while (filterItems.hasNext()) {
            term = (String)filterItems.next();
            
            if (term.equals("[and]")) {
                op = checkOp(op, Operation.CONJUNCTION);
            } else if (term.equals("[or]")) {
                op = checkOp(op, Operation.DISJUNCTION);
            } else if (term.equals("[)]")){
                break;
            } else
                throw new IllegalArgumentException("Unexpected token '" + term + "'");

            operands.add(parseFilter(query, filterItems));
        }
        
        if (op == null)
            return (CqQuery.Filter) operands.get(0);
        
        return query.cqProvider().buildFilterNode(op, (CqQuery.Filter[]) operands
            .toArray(new CqQuery.Filter[operands.size()]));
    }

    /**
     * Parses a specification for a query display field and constructs a
     * DisplayField object to represent it. The specification takes the general
     * form:<br>
     * <br>
     * <b>[</b><i>sort-key-position</i><b>]</b><i>field-path</i><b>{</b><i>label</i><b>}</b><br>
     * <br>
     * Both <b>[</b><i>sort-key-position</i><b>]</b> and the {</b><i>label</i><b>}</b>
     * are optional. The first position in the sort key is index 1. The index
     * may be preceded by <b>D</b> to indicate a descending sort. Enclose the
     * entire specification in square brackets (<b>[]</b>)to make the field not visible.
     * 
     * @param query A CqQuery proxy for the query for which the DisplayField is
     *            to be constructed. Must define the properties required by
     *            buildFieldPath.
     * @param fieldSpec A String containing the specification for one display
     *            field of a query.
     * @return A DisplayField object implementing the String specification.
     * @throws WvcmException If the necessary information is not available or
     *             obtainable from the server.
     */
    protected static DisplayField parseDisplayField(CqQuery query,
                                                    String fieldSpec)
        throws WvcmException
    {
        boolean isVisible = true;
        long sortOrder = 0;
        SortType sortType = SortType.NO_SORT;
        String label = null;

        // [field] => invisible field
        if (fieldSpec.startsWith("[")
            && fieldSpec.endsWith("]")) {
            isVisible = false;
            fieldSpec = fieldSpec.substring(1, fieldSpec.length() - 1);
        }

        // [n]field ==> position n in sort key; [Dn]field ==> descending sort
        if (fieldSpec.startsWith("[")) {
            int rb = fieldSpec.indexOf("]");
            String sortOrderSpec = fieldSpec.substring(1, rb);

            fieldSpec = fieldSpec.substring(rb + 1);

            if (sortOrderSpec.startsWith("D")) {
                sortType = CqQuery.DisplayField.SortType.DESCENDING;
                sortOrderSpec = sortOrderSpec.substring(1);
            } else {
                sortType = CqQuery.DisplayField.SortType.ASCENDING;
            }

            try {
                sortOrder = Integer.parseInt(sortOrderSpec);
            } catch (Exception ex) {
                throw new IllegalArgumentException("Cannot interpret '"
                    + sortOrderSpec + "' as a sort specification in '"
                    + fieldSpec + "'");
            }
        }

        // field{label} ==> alternate specification for field label
        if (fieldSpec.endsWith("}")) {
            int lb = fieldSpec.lastIndexOf("{");

            label = fieldSpec.substring(lb + 1, fieldSpec.length() - 1);
            fieldSpec = fieldSpec.substring(0, lb);
        } else {
            label = fieldSpec;
        }

        CqQuery.DisplayField field = query.cqProvider().buildDisplayField();

        field.setIsVisible(isVisible);
        field.setSortOrder(sortOrder);
        field.setSortType(sortType);
        field.setLabel(label);
        field.setPath(buildFieldPath(query, fieldSpec));

        // TODO Aggregates
        // TODO Functions
        // TODO Group By

        return field;
    }

   /** 
    * Properties needed from a record type of a database 
    * (without its field definitions) 
    */
   private static final PropertyRequest RECORD_TYPE_WITHOUT_FIELDS = 
       new PropertyRequest(
            CqRecordType.DISPLAY_NAME,
            CqRecordType.USER_FRIENDLY_LOCATION);

   /** CqFieldDefinition properties we will be needing */
   private static final PropertyRequest FIELD_DEFINITION_PROPERTIES = 
       new PropertyRequest(
            CqFieldDefinition.DISPLAY_NAME,
            CqFieldDefinition.RECORD_TYPE.nest(RECORD_TYPE_WITHOUT_FIELDS),
            CqFieldDefinition.VALUE_TYPE,
            CqFieldDefinition.FIELD_TYPE,
            CqFieldDefinition.REFERENCED_RECORD_TYPE
                .nest(RECORD_TYPE_WITHOUT_FIELDS),
            CqFieldDefinition.USER_FRIENDLY_LOCATION);

   /** Properties needed from a record type of a database */
   protected static final PropertyRequest RECORD_TYPE_WITH_FIELDS =
       new PropertyRequest(
            CqRecordType.DISPLAY_NAME,
            CqRecordType.USER_FRIENDLY_LOCATION,
            CqRecordType.FIELD_DEFINITIONS.nest(FIELD_DEFINITION_PROPERTIES)
        );
}