/*
 * file Utilities.java
 *
 * Licensed Materials - Property of IBM
 * Restricted Materials of IBM - you are allowed to copy, modify and 
 * redistribute this file as part of any program that interfaces with 
 * IBM Rational CM API.
 *
 * com.ibm.rational.stp.client.samples.Utilities
 *
 *  Copyright IBM Corporation 2004, 2008.  All Rights Reserved.
 * Note to U.S. Government Users Restricted Rights:  Use, duplication or 
 * disclosure restricted by GSA ADP  Schedule Contract with IBM Corp.
 */

package com.ibm.rational.stp.client.samples;

import java.awt.Component;
import java.io.File;
import java.io.FileReader;
import java.lang.reflect.InvocationTargetException;
import java.util.List;

import javax.swing.JOptionPane;
import javax.wvcm.Feedback;
import javax.wvcm.ProviderFactory;
import javax.wvcm.Resource;
import javax.wvcm.ResourceList;
import javax.wvcm.WvcmException;
import javax.wvcm.PropertyRequestItem.PropertyRequest;
import javax.wvcm.ProviderFactory.Callback;
import javax.wvcm.ProviderFactory.Callback.Authentication;

import com.ibm.rational.wvcm.stp.StpException;
import com.ibm.rational.wvcm.stp.StpProvider;
import com.ibm.rational.wvcm.stp.StpResource;
import com.ibm.rational.wvcm.stp.StpProvider.Domain;
import com.ibm.rational.wvcm.stp.StpProvider.StpCallback;
import com.ibm.rational.wvcm.stp.cq.CqDbSet;
import com.ibm.rational.wvcm.stp.cq.CqProvider;
import com.ibm.rational.wvcm.stp.cq.CqUser;
import com.ibm.rational.wvcm.stp.cq.CqUserDb;

/**
 * A collection of static utility methods supporting the CM API examples
 */
public class Utilities {

    /**
     * A simple Authentication object in which the username and password
     * obtained from the user is cached for use by the CM API.
     */
    static class UnPw implements Authentication {
        /**
         * Constructs an Authentication object
         * 
         * @param unpw A String[] containing the username and password.
         */
        UnPw(String[] unpw) {  m_data = unpw; }

        public String loginName() { return m_data[0]; }
        public String password() { return m_data.length > 1 ? m_data[1] : ""; };

        /** The cached credentials */
        private String[] m_data;
    }

    /** Password used by static callback */
    public static String g_pass = "";
    /** User name used by static callback */
    public static String g_user = "admin";

    private static Callback g_callback = new Callback()
    {
        public Authentication getAuthentication(String realm, int retryCount)
        {
            if (retryCount > 0)
                throw new IllegalAccessError("Wrong username/password for " + realm);
            return new Authentication()
            {
                public String loginName()
                {
                    return g_user;
                }

                public String password()
                {
                    return g_pass;
                }
            };
        }
    };

    public static CqProvider getStaticProvider()
    throws Exception
    {
        return (CqProvider) ProviderFactory
            .createProvider(CqProvider.CQ_ONLY_PROVIDER_CLASS, g_callback);        
    }
    
    /**
     * Constructs an instance of the CM API provider with or without an
     * authenticator.
     * 
     * @return The instantiated Provider object
     * @throws Exception
     *             If the Provider could not be instantiated
     */
    static StpProvider getProvider() throws Exception {
        try {
            Callback callback = new StpCallback() {
                private UnPw m_unpw;
    
                public Authentication getAuthentication(String r, int c)
                {   return null; /* Will not be called */   }
    
                public Authentication getAuthenticationEx(Domain domain,
                                                          String realm,
                                                          int retryCount,
                                                          StpProvider provider,
                                                          WvcmException failure) 
                throws WvcmException
                {
                    // Try to reuse last credentials on each new repository
                    if (m_unpw != null && retryCount == 0)
                        return m_unpw;
    
                    String title = "Enter " + domain 
                            + " Username '+' Password for "
                            + realm + " [" + retryCount + "]";
                    
                    if (failure != null)
                        title = "Login failed: " + failure + "\n" + title;
                    
                    String unpw = JOptionPane.showInputDialog(title, "admin+");
    
                    if (unpw == null || unpw.length() == 0)
                        throw new IllegalAccessError("User canceled request");
                    
                    if (unpw.equals("anonymous"))
                        return null;
                    
                    if (unpw.startsWith("@")) {
                        File file = new File(unpw.substring(1));
    
                        try {
                            FileReader reader = new FileReader(file);
                            char[] buf = new char[100];
                            int count = reader.read(buf);
    
                            unpw = new String(buf, 0, count);
                            reader.close();
                        } catch (Throwable t) {
                            Utilities.exception(null,
                                                "Reading password file " + unpw,
                                                t);
                        }
                    }
    
                    return m_unpw = new UnPw(unpw.split("\\+", -2));
                }
            };
    
            // Instantiate a Provider
            return (StpProvider) ProviderFactory
                .createProvider(StpProvider.PROVIDER_CLASS, callback);
        } catch (InvocationTargetException ite) {
            WvcmException e = (WvcmException) ite.getTargetException();
    
            System.out.println("*** " + e);
            
            for (Throwable nested: e.getNestedExceptions())
                System.out.println("***  " + nested);
    
            throw e;
        }
    }

    static ResourceList<CqUserDb> getUserDbList(CqProvider provider, 
                                                PropertyRequest feedback)
    throws WvcmException
    {
        PropertyRequest wantedProps =
            new PropertyRequest(CqDbSet.CURRENT_USER
                .nest(CqUser.SUBSCRIBED_DATABASES.nest(feedback)));
        ResourceList<CqUserDb> result = provider.resourceList();

        for (CqDbSet set : provider.doGetDbSetList(wantedProps)) {
            if (set.getResourceError() == null)
                result.addAll(set.getCurrentUser().getSubscribedDatabases());
        }

        return result;
    }
    
    /**
     * Extracts the message content from a Throwable and returns it as a
     * hierarchical array of Strings capturing the nesting of the Throwable's
     * message components. This structure formats reasonably in a SWING
     * showMessageDialog invocation.
     * 
     * @param ex The Throwable object whose message content is to be extracted.
     * @return If the given Throwable has nested components, an array consisting
     *         of the Throwable's message and an array of the nested messages.
     */
    private static Object messages(Throwable ex) {
        String msg = ex.getLocalizedMessage();
    
        if (msg == null || msg.length() == 0)
            msg = ex.toString();
    
        if (ex instanceof StpException) {
            Throwable[] nested = ((StpException) ex).getNestedExceptions();
    
            if (nested != null && nested.length > 0) {
                Object[] msgs = new Object[nested.length];
    
                for (int i = 0; i < msgs.length; ++i)
                    msgs[i] = messages(nested[i]);
    
                return new Object[] { msg, msgs };
            }
        } else if (ex.getCause() != null) {
            return new Object[] {msg, new Object[]{messages(ex.getCause())}};
        }
    
        return msg;
    }
    
    /**
     * Displays a Swing dialog containing the messages associated with a given
     * Throwable.
     * 
     * @param frame The parent frame for the message dialog.
     * @param title The title to appear in the dialog window.
     * @param ex The throwable whose messages are to be displayed.
     */
    static void exception(Component frame, String title, Throwable ex) {
        JOptionPane.showMessageDialog(frame,
                                      messages(ex),
                                      title,
                                      JOptionPane.ERROR_MESSAGE);
    }

    /**
     * Returns a String suitable for displaying the type of resource as
     * determined by its proxy class.
     * 
     * @param resource
     *            A Resource proxy from which the type of resource will be
     *            determined.
     * @return A String containing the simple name of the most derived CM API 
     * interface implemented by the given proxy.
     */
    static String resourceType(Resource resource) {
        Class interfaces[] = resource.getClass().getInterfaces();
        Class<?> choice = Resource.class;

        for (int i = 0; i < interfaces.length; ++i) {
            String name = interfaces[i].getName();
            if (name.startsWith("com.ibm.rational.wvcm.")
                || name.startsWith("javax.wvcm")) {

                // Within the CM API, select the most derived interface
                // this resource implements
                if (choice.isAssignableFrom(interfaces[i]))
                    choice = interfaces[i];
            }
        }

        String name = choice.getName();
        
        return name.substring(1+name.lastIndexOf('.'));
    }
    
    static boolean isListOfResources(Object val)
    {
        return val instanceof ResourceList;
    }
    
    static ResourceList<? extends StpResource> toResourceList(StpProvider provider, 
                                                              List list)
    {
        if (list instanceof ResourceList)
            return (ResourceList<? extends StpResource>)list;
        
        return provider.resourceList();
    }
}
